# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.commands import CliCommandType

from azure.cli.command_modules.batch import _client_factory as factories
from azure.cli.command_modules.batch._validators import (
    validate_pool_settings, validate_cert_settings)
from azure.cli.command_modules.batch._exception_handler import batch_exception_handler
from azure.cli.command_modules.batch._format import (
    job_list_table_format,
    task_create_table_format,
    account_keys_list_table_format,
    account_list_table_format,
    application_list_table_format,
    account_keys_renew_table_format)


def _operation(name):
    return "".join(x.title() for x in name.split('_')) + "Operations"


# pylint: disable=too-many-locals, too-many-statements, line-too-long
def load_command_table(self, _):

    data_path = 'azure.batch.operations._{}_operations#{}.'
    mgmt_path = 'azure.mgmt.batch.operations._{}_operations#{}.'

    def get_data_type(name):
        return CliCommandType(
            operations_tmpl=data_path.format(name, _operation(name)) + "{}",
            client_factory=get_data_factory(name),
            exception_handler=batch_exception_handler
        )

    def get_mgmt_type(name):
        return CliCommandType(
            operations_tmpl=mgmt_path.format(name, _operation(name)) + "{}",
            client_factory=get_mgmt_factory(name),
            exception_handler=batch_exception_handler
        )

    def get_mgmt_factory(name):
        return getattr(factories, "mgmt_{}_client_factory".format(name))

    def get_data_factory(name):
        return getattr(factories, "{}_client_factory".format(name))

    # Mgmt Account Operations
    with self.command_group('batch account', get_mgmt_type('batch_account'), client_factory=get_mgmt_factory('batch_account')) as g:
        g.custom_command('list', 'list_accounts', table_transformer=account_list_table_format)
        g.custom_show_command('show', 'get_account')
        g.custom_command('create', 'create_account', supports_no_wait=True)
        g.custom_command('set', 'update_account')
        g.command('delete', 'delete', supports_no_wait=True, confirmation=True)
        g.custom_command('login', 'login_account')
        g.command('autostorage-keys sync', 'synchronize_auto_storage_keys')
        g.command('keys list', 'get_keys', table_transformer=account_keys_list_table_format)
        g.command('keys renew', 'regenerate_key', table_transformer=account_keys_renew_table_format)

    with self.command_group('batch application', get_mgmt_type('application'), client_factory=get_mgmt_factory('application')) as g:
        g.command('list', 'list', table_transformer=application_list_table_format)
        g.show_command('show', 'get')
        g.command('create', 'create')
        g.custom_command('set', 'update_application')
        g.command('delete', 'delete', confirmation=True)

    with self.command_group('batch application package', get_mgmt_type('application_package'), client_factory=get_mgmt_factory('application_package'))as g:
        g.custom_command('create', 'create_application_package')
        g.command('delete', 'delete', confirmation=True)
        g.show_command('show', 'get')
        g.command('activate', 'activate')
        g.command('list', 'list')

    with self.command_group('batch location quotas', get_mgmt_type('location')) as g:
        g.show_command('show', 'get_quotas')

    # Data Plane Commands
    with self.command_group('batch application summary', get_data_type('application')) as g:
        g.batch_command('list', 'list')
        g.batch_command('show', 'get')

    with self.command_group('batch pool supported-images', get_data_type('account')) as g:
        g.batch_command('list', 'list_supported_images')

    with self.command_group('batch pool node-counts', get_data_type('account')) as g:
        g.batch_command('list', 'list_pool_node_counts')

    with self.command_group('batch certificate', get_data_type('certificate'), client_factory=get_data_factory('certificate')) as g:
        g.custom_command('create', 'create_certificate')
        g.custom_command('delete', 'delete_certificate', confirmation=True)
        g.batch_command('show', 'get', validator=validate_cert_settings)
        g.batch_command('list', 'list')

    pool_type = get_data_type('pool')
    with self.command_group('batch pool', pool_type, client_factory=get_data_factory('pool')) as g:
        g.batch_command('usage-metrics list', 'list_usage_metrics')
        g.batch_command('all-statistics show', 'get_all_lifetime_statistics')
        g.batch_command('create', 'add', validator=validate_pool_settings, flatten=10)
        g.batch_command('list', 'list')
        g.batch_command('delete', 'delete')
        g.batch_command('show', 'get')
        g.batch_command('set', 'patch')
        g.custom_command('reset', 'update_pool')
        g.batch_command('autoscale disable', 'disable_auto_scale')
        g.batch_command('autoscale enable', 'enable_auto_scale')
        g.batch_command('autoscale evaluate', 'evaluate_auto_scale')
        g.custom_command('resize', 'resize_pool')

    with self.command_group('batch node', pool_type) as g:
        g.batch_command('delete', 'remove_nodes')

    with self.command_group('batch job', get_data_type('job'), client_factory=get_data_factory('job')) as g:
        g.batch_command('all-statistics show', 'get_all_lifetime_statistics')
        g.batch_command('create', 'add')
        g.batch_command('delete', 'delete')
        g.batch_command('show', 'get')
        g.batch_command('set', 'patch', flatten=2)
        g.batch_command('reset', 'update', flatten=2)
        g.custom_command('list', 'list_job', table_transformer=job_list_table_format)
        g.batch_command('disable', 'disable')
        g.batch_command('enable', 'enable')
        g.batch_command('stop', 'terminate')
        g.batch_command('prep-release-status list', 'list_preparation_and_release_task_status')
        g.batch_command('task-counts show', 'get_task_counts')

    with self.command_group('batch job-schedule', get_data_type('job_schedule')) as g:
        g.batch_command('create', 'add')
        g.batch_command('delete', 'delete')
        g.batch_command('show', 'get')
        g.batch_command('set', 'patch')
        g.batch_command('reset', 'update')
        g.batch_command('disable', 'disable')
        g.batch_command('enable', 'enable')
        g.batch_command('stop', 'terminate')
        g.batch_command('list', 'list')

    with self.command_group('batch task', get_data_type('task'), client_factory=get_data_factory('task')) as g:
        g.custom_command('create', 'create_task', table_transformer=task_create_table_format)
        g.batch_command('list', 'list')
        g.batch_command('delete', 'delete')
        g.batch_command('show', 'get')
        g.batch_command('reset', 'update')
        g.batch_command('reactivate', 'reactivate')
        g.batch_command('stop', 'terminate')
        g.batch_command('subtask list', 'list_subtasks')

    file_type = get_data_type('file')
    with self.command_group('batch task file', file_type) as g:
        g.batch_command('delete', 'delete_from_task')
        g.batch_command('download', 'get_from_task')
        g.batch_command('show', 'get_properties_from_task')
        g.batch_command('list', 'list_from_task')

    with self.command_group('batch node file', file_type) as g:
        g.batch_command('delete', 'delete_from_compute_node')
        g.batch_command('download', 'get_from_compute_node')
        g.batch_command('show', 'get_properties_from_compute_node')
        g.batch_command('list', 'list_from_compute_node')

    with self.command_group('batch node', get_data_type('compute_node')) as g:
        g.batch_command('user create', 'add_user')
        g.batch_command('user delete', 'delete_user')
        g.batch_command('user reset', 'update_user')
        g.batch_command('show', 'get')
        g.batch_command('list', 'list')
        g.batch_command('reboot', 'reboot')
        g.batch_command('reimage', 'reimage')
        g.batch_command('scheduling disable', 'disable_scheduling')
        g.batch_command('scheduling enable', 'enable_scheduling')
        g.batch_command('remote-login-settings show', 'get_remote_login_settings')
        g.batch_command('remote-desktop download', 'get_remote_desktop')
        g.batch_command('service-logs upload', 'upload_batch_service_logs')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import datetime

from azure.cli.core.util import CLIError, sdk_no_wait
from azure.cli.core.commands import LongRunningOperation
from azure.cli.core.profiles import ResourceType
from knack.log import get_logger
from ._client_factory import (resource_client_factory)

logger = get_logger(__name__)


def validate_and_deploy_arm_template(cmd, resource_group_name, template, parameters):
    suffix = datetime.datetime.now().strftime("%Y%m%d%H%M")
    deployment_name = 'AzurePSDeployment-' + suffix

    logger.info("Validating the deployment")
    validate_result = _deploy_arm_template_core(
        cmd, resource_group_name, template, parameters, deployment_name, 'incremental', True)
    if validate_result.error is not None:
        errors_detailed = _build_detailed_error(validate_result.error, [])
        errors_detailed.insert(0, "Error validating template. See below for more information.")
        raise CLIError('\n'.join(errors_detailed))
    logger.info("Deployment is valid, and begin to deploy")
    _deploy_arm_template_core(cmd, resource_group_name, template,
                              parameters, deployment_name, 'incremental', False)


def _deploy_arm_template_core(cmd,
                              resource_group_name,
                              template,
                              parameters,
                              deployment_name=None,
                              mode='incremental',
                              validate_only=False,
                              no_wait=False):
    DeploymentProperties = cmd.get_models('DeploymentProperties', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES)
    properties = DeploymentProperties(
        template=template, template_link=None, parameters=parameters, mode=mode)
    client = resource_client_factory(cmd.cli_ctx)

    if cmd.supported_api_version(min_api='2019-10-01', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES):
        Deployment = cmd.get_models('Deployment', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES)
        deployment = Deployment(properties=properties)

        if validate_only:
            deploy_poll = sdk_no_wait(no_wait, client.deployments.validate, resource_group_name, deployment_name,
                                      deployment)
        else:
            deploy_poll = sdk_no_wait(no_wait, client.deployments.create_or_update, resource_group_name,
                                      deployment_name, deployment)
        return LongRunningOperation(cmd.cli_ctx)(deploy_poll)

    if validate_only:
        deploy_poll = sdk_no_wait(no_wait, client.deployments.validate, resource_group_name, deployment_name,
                                  properties)
    else:
        deploy_poll = sdk_no_wait(no_wait, client.deployments.create_or_update, resource_group_name, deployment_name,
                                  properties)
    return LongRunningOperation(cmd.cli_ctx)(deploy_poll)


def _build_detailed_error(top_error, output_list):
    if output_list:
        output_list.append(' Inner Error - Code: "{}" Message: "{}"'.format(top_error.code, top_error.message))
    else:
        output_list.append('Error - Code: "{}" Message: "{}"'.format(top_error.code, top_error.message))

    if top_error.details:
        for error in top_error.details:
            _build_detailed_error(error, output_list)

    return output_list



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


def cf_media(cli_ctx, *_):
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    from azure.mgmt.media import AzureMediaServices
    return get_mgmt_service_client(cli_ctx, AzureMediaServices)


def get_mediaservices_client(cli_ctx, *_):
    return cf_media(cli_ctx).mediaservices


def _auth_client_factory(cli_ctx, scope=None):
    import re
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    from azure.cli.core.profiles import ResourceType
    subscription_id = None
    if scope:
        matched = re.match('/subscriptions/(?P<subscription>[^/]*)/', scope)
        if matched:
            subscription_id = matched.groupdict()['subscription']
    return get_mgmt_service_client(cli_ctx, ResourceType.MGMT_AUTHORIZATION, subscription_id=subscription_id)


def _graph_client_factory(cli_ctx, **_):
    from azure.cli.core._profile import Profile
    from azure.cli.core.commands.client_factory import configure_common_settings
    from azure.graphrbac import GraphRbacManagementClient
    profile = Profile(cli_ctx=cli_ctx)
    cred, _, tenant_id = profile.get_login_credentials(
        resource=cli_ctx.cloud.endpoints.active_directory_graph_resource_id)
    client = GraphRbacManagementClient(cred, tenant_id,
                                       base_url=cli_ctx.cloud.endpoints.active_directory_graph_resource_id)
    configure_common_settings(cli_ctx, client)
    return client


def get_transforms_client(cli_ctx, *_):
    return cf_media(cli_ctx).transforms


def get_assets_client(cli_ctx, *_):
    return cf_media(cli_ctx).assets


def get_jobs_client(cli_ctx, *_):
    return cf_media(cli_ctx).jobs


def get_streaming_locators_client(cli_ctx, *_):
    return cf_media(cli_ctx).streaming_locators


def get_streaming_policies_client(cli_ctx, *_):
    return cf_media(cli_ctx).streaming_policies


def get_streaming_endpoints_client(cli_ctx, *_):
    return cf_media(cli_ctx).streaming_endpoints


def get_locations_client(cli_ctx, *_):
    return cf_media(cli_ctx).locations


def get_live_events_client(cli_ctx, *_):
    return cf_media(cli_ctx).live_events


def get_live_outputs_client(cli_ctx, *_):
    return cf_media(cli_ctx).live_outputs


def get_content_key_policies_client(cli_ctx, *_):
    return cf_media(cli_ctx).content_key_policies


def get_account_filters_client(cli_ctx, *_):
    return cf_media(cli_ctx).account_filters


def get_asset_filters_client(cli_ctx, *_):
    return cf_media(cli_ctx).asset_filters



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=line-too-long

from azure.cli.core.commands import CliCommandType
from azure.cli.core.util import empty_on_404

from ._client_factory import (
    cf_signalr)


def load_command_table(self, _):

    signalr_custom_utils = CliCommandType(
        operations_tmpl='azure.cli.command_modules.signalr.custom#{}',
        client_factory=cf_signalr
    )

    signalr_key_utils = CliCommandType(
        operations_tmpl='azure.cli.command_modules.signalr.key#{}',
        client_factory=cf_signalr
    )

    signalr_cors_utils = CliCommandType(
        operations_tmpl='azure.cli.command_modules.signalr.cors#{}',
        client_factory=cf_signalr
    )

    signalr_network_utils = CliCommandType(
        operations_tmpl='azure.cli.command_modules.signalr.network#{}',
        client_factory=cf_signalr
    )

    signalr_upstream_utils = CliCommandType(
        operations_tmpl='azure.cli.command_modules.signalr.upstream#{}',
        client_factory=cf_signalr
    )

    with self.command_group('signalr', signalr_custom_utils) as g:
        g.command('create', 'signalr_create')
        g.command('delete', 'signalr_delete')
        g.command('list', 'signalr_list')
        g.show_command('show', 'signalr_show', exception_handler=empty_on_404)
        g.command('restart', 'signalr_restart', exception_handler=empty_on_404)
        g.generic_update_command('update', getter_name='signalr_update_get',
                                 setter_name='signalr_update_set', custom_func_name='signalr_update_custom',
                                 custom_func_type=signalr_custom_utils)

    with self.command_group('signalr key', signalr_key_utils) as g:
        g.command('list', 'signalr_key_list')
        g.command('renew', 'signalr_key_renew')

    with self.command_group('signalr cors', signalr_cors_utils) as g:
        g.command('add', 'signalr_cors_add')
        g.command('remove', 'signalr_cors_remove')
        g.command('list', 'signalr_cors_list')

    with self.command_group('signalr network-rule', signalr_network_utils) as g:
        g.command('list', 'list_network_rules')
        g.command('update', 'update_network_rules')

    with self.command_group('signalr upstream', signalr_upstream_utils) as g:
        g.command('list', 'signalr_upstream_list')
        g.command('update', 'signalr_upstream_update')
        g.command('clear', 'signalr_upstream_clear')



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------
# pylint: disable=line-too-long
# pylint: disable=too-many-lines
# pylint: disable=too-many-statements

from azure.cli.core.commands.parameters import (
    tags_type,
    get_enum_type,
    resource_group_name_type,
    get_location_type
)
from azure.cli.core.commands.validators import get_default_location_from_resource_group
from ..action import AddSku


def load_arguments(self, _):

    with self.argument_context('databoxedge device list') as c:
        c.argument('resource_group_name', resource_group_name_type)
        c.argument('expand', type=str, help='Specify $expand=details to populate additional fields related to the '
                   'resource or Specify $skipToken=<token> to populate the next page in the list.')

    with self.argument_context('databoxedge device show') as c:
        c.argument('device_name', options_list=['--name', '-n', '--device-name'], type=str, help='The device name.',
                   id_part='name')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge device create') as c:
        c.argument('device_name', options_list=['--name', '-n', '--device-name'], type=str, help='The device name.')
        c.argument('resource_group_name', resource_group_name_type)
        c.argument('location', arg_type=get_location_type(self.cli_ctx), required=False,
                   validator=get_default_location_from_resource_group)
        c.argument('tags', tags_type)
        c.argument('sku', action=AddSku, nargs='+', help='The SKU type.')
        c.argument('etag', type=str, help='The etag for the devices.')
        c.argument('data_box_edge_device_status', options_list=['--status'], arg_type=get_enum_type(['ReadyToSetup',
                                                                                                     'Online',
                                                                                                     'Offline',
                                                                                                     'NeedsAttention',
                                                                                                     'Disconnected',
                                                                                                     'PartiallyDisconnected',
                                                                                                     'Maintenance']),
                   help='The status of the Data Box Edge/Gateway device.')
        c.argument('description', type=str, help='The Description of the Data Box Edge/Gateway device.')
        c.argument('model_description', type=str, help='The description of the Data Box Edge/Gateway device model.')
        c.argument('friendly_name', type=str, help='The Data Box Edge/Gateway device name.')

    with self.argument_context('databoxedge device update') as c:
        c.argument('device_name', options_list=['--name', '-n', '--device-name'], type=str, help='The device name.',
                   id_part='name')
        c.argument('resource_group_name', resource_group_name_type)
        c.argument('tags', tags_type)

    with self.argument_context('databoxedge device delete') as c:
        c.argument('device_name', options_list=['--name', '-n', '--device-name'], type=str, help='The device name.',
                   id_part='name')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge device download-update') as c:
        c.argument('device_name', options_list=['--name', '-n', '--device-name'], type=str, help='The device name.',
                   id_part='name')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge device install-update') as c:
        c.argument('device_name', options_list=['--name', '-n', '--device-name'], type=str, help='The device name.',
                   id_part='name')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge device scan-for-update') as c:
        c.argument('device_name', options_list=['--name', '-n', '--device-name'], type=str, help='The device name.',
                   id_part='name')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge device show-update-summary') as c:
        c.argument('device_name', options_list=['--name', '-n', '--device-name'], type=str, help='The device name.',
                   id_part='name')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge device wait') as c:
        c.argument('device_name', options_list=['--name', '-n', '--device-name'], type=str, help='The device name.',
                   id_part='name')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge alert list') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge alert show') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.',
                   id_part='name')
        c.argument('name', options_list=['--name', '-n'], type=str, help='The alert name.', id_part='child_name_1')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge bandwidth-schedule list') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge bandwidth-schedule show') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.',
                   id_part='name')
        c.argument('name', options_list=['--name', '-n'], type=str, help='The bandwidth schedule name.',
                   id_part='child_name_1')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge bandwidth-schedule create') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.')
        c.argument('name', options_list=['--name', '-n'], type=str, help='The bandwidth schedule name which needs to '
                   'be added/updated.')
        c.argument('resource_group_name', resource_group_name_type)
        c.argument('start', type=str, help='The start time of the schedule in UTC.')
        c.argument('stop', type=str, help='The stop time of the schedule in UTC.')
        c.argument('rate_in_mbps', type=int, help='The bandwidth rate in Mbps.')
        c.argument('days', nargs='+', help='The days of the week when this schedule is applicable.')

    with self.argument_context('databoxedge bandwidth-schedule update') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.',
                   id_part='name')
        c.argument('name', options_list=['--name', '-n'], type=str, help='The bandwidth schedule name which needs to '
                   'be added/updated.', id_part='child_name_1')
        c.argument('resource_group_name', resource_group_name_type)
        c.argument('start', type=str, help='The start time of the schedule in UTC.')
        c.argument('stop', type=str, help='The stop time of the schedule in UTC.')
        c.argument('rate_in_mbps', type=int, help='The bandwidth rate in Mbps.')
        c.argument('days', nargs='+', help='The days of the week when this schedule is applicable.')
        c.ignore('parameters')

    with self.argument_context('databoxedge bandwidth-schedule delete') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.',
                   id_part='name')
        c.argument('name', options_list=['--name', '-n'], type=str, help='The bandwidth schedule name.',
                   id_part='child_name_1')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge bandwidth-schedule wait') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.',
                   id_part='name')
        c.argument('name', options_list=['--name', '-n'], type=str, help='The bandwidth schedule name.',
                   id_part='child_name_1')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge show-job') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.',
                   id_part='name')
        c.argument('name', options_list=['--name', '-n'], type=str, help='The job name.', id_part='child_name_1')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge list-node') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge order list') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge order show') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.',
                   id_part='name')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge order create') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str,
                   help='The order details of a device.')
        c.argument('resource_group_name', resource_group_name_type)
        c.argument('status', arg_type=get_enum_type(['Untracked', 'AwaitingFulfilment', 'AwaitingPreparation',
                                                     'AwaitingShipment', 'Shipped', 'Arriving', 'Delivered',
                                                     'ReplacementRequested', 'LostDevice', 'Declined',
                                                     'ReturnInitiated', 'AwaitingReturnShipment', 'ShippedBack',
                                                     'CollectedAtMicrosoft']), help='Status of the order as per the '
                   'allowed status types.', arg_group='Current Status')
        c.argument('comments', type=str, help='Comments related to this status change.', arg_group='Current Status')
        c.argument('address_line1', type=str, help='The address line1.', arg_group='Shipping Address')
        c.argument('address_line2', type=str, help='The address line2.', arg_group='Shipping Address')
        c.argument('address_line3', type=str, help='The address line3.', arg_group='Shipping Address')
        c.argument('postal_code', type=str, help='The postal code.', arg_group='Shipping Address')
        c.argument('city', type=str, help='The city name.', arg_group='Shipping Address')
        c.argument('state', type=str, help='The state name.', arg_group='Shipping Address')
        c.argument('country', type=str, help='The country name.', arg_group='Shipping Address')
        c.argument('contact_person', type=str, help='The contact person name.', arg_group='Contact Information')
        c.argument('company_name', type=str, help='The name of the company.', arg_group='Contact Information')
        c.argument('phone', type=str, help='The phone number.', arg_group='Contact Information')
        c.argument('email_list', nargs='+', help='The email list.', arg_group='Contact Information')

    with self.argument_context('databoxedge order update') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str,
                   help='The order details of a device.', id_part='name')
        c.argument('resource_group_name', resource_group_name_type)
        c.argument('status', arg_type=get_enum_type(['Untracked', 'AwaitingFulfilment', 'AwaitingPreparation',
                                                     'AwaitingShipment', 'Shipped', 'Arriving', 'Delivered',
                                                     'ReplacementRequested', 'LostDevice', 'Declined',
                                                     'ReturnInitiated', 'AwaitingReturnShipment', 'ShippedBack',
                                                     'CollectedAtMicrosoft']), help='Status of the order as per the '
                   'allowed status types.', arg_group='Current Status')
        c.argument('comments', type=str, help='Comments related to this status change.', arg_group='Current Status')
        c.argument('address_line1', type=str, help='The address line1.', arg_group='Shipping Address')
        c.argument('address_line2', type=str, help='The address line2.', arg_group='Shipping Address')
        c.argument('address_line3', type=str, help='The address line3.', arg_group='Shipping Address')
        c.argument('postal_code', type=str, help='The postal code.', arg_group='Shipping Address')
        c.argument('city', type=str, help='The city name.', arg_group='Shipping Address')
        c.argument('state', type=str, help='The state name.', arg_group='Shipping Address')
        c.argument('country', type=str, help='The country name.', arg_group='Shipping Address')
        c.argument('contact_person', type=str, help='The contact person name.', arg_group='Contact Information')
        c.argument('company_name', type=str, help='The name of the company.', arg_group='Contact Information')
        c.argument('phone', type=str, help='The phone number.', arg_group='Contact Information')
        c.argument('email_list', nargs='+', help='The email list.', arg_group='Contact Information')
        c.ignore('order')

    with self.argument_context('databoxedge order delete') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.',
                   id_part='name')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge order wait') as c:
        c.argument('device_name', options_list=['--device-name', '-d'], type=str, help='The device name.',
                   id_part='name')
        c.argument('resource_group_name', resource_group_name_type)

    with self.argument_context('databoxedge list-sku') as c:
        c.argument('filter_', options_list=['--filter'], type=str, help='Specify $filter=\'location eq <location>\' to '
                   'filter on location.')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from enum import Enum
import json
from knack.log import get_logger
from azure.cli.core.util import shell_safe_json_parse
from ._models import KeyValue
from ._constants import FeatureFlagConstants

# pylint: disable=too-few-public-methods
# pylint: disable=too-many-instance-attributes

logger = get_logger(__name__)

# Feature Flag Models #


class FeatureState(Enum):
    OFF = 1
    ON = 2
    CONDITIONAL = 3


class FeatureQueryFields(Enum):
    KEY = 0x001
    LABEL = 0x002
    LAST_MODIFIED = 0x020
    LOCKED = 0x040
    STATE = 0x100
    DESCRIPTION = 0x200
    CONDITIONS = 0x400
    ALL = KEY | LABEL | LAST_MODIFIED | LOCKED | STATE | DESCRIPTION | CONDITIONS


class FeatureFlagValue:
    '''
    Schema of Value inside KeyValue when key is a Feature Flag.

    :ivar str id:
        ID (key) of the feature.
    :ivar str description:
        Description of Feature Flag
    :ivar bool enabled:
        Represents if the Feature flag is On/Off/Conditionally On
    :ivar dict {string, FeatureFilter[]} conditions:
        Dictionary that contains client_filters List (and server_filters List in future)
    '''

    def __init__(self,
                 id_,
                 description=None,
                 enabled=None,
                 conditions=None):
        default_conditions = {'client_filters': []}

        self.id = id_
        self.description = description
        self.enabled = enabled if enabled else False
        self.conditions = conditions if conditions else default_conditions

    def __repr__(self):
        featureflagvalue = {
            "id": self.id,
            "description": self.description,
            "enabled": self.enabled,
            "conditions": custom_serialize_conditions(self.conditions)
        }

        return json.dumps(featureflagvalue, indent=2, ensure_ascii=False)


class FeatureFlag:
    '''
    Feature Flag schema as displayed to the user.

    :ivar str key:
        FeatureName (key) of the entry.
    :ivar str label:
        Label of the entry.
    :ivar str state:
        Represents if the Feature flag is On/Off/Conditionally On
    :ivar str description:
        Description of Feature Flag
    :ivar bool locked:
        Represents whether the feature flag is locked.
    :ivar str last_modified:
        A str representation of the datetime object representing the last time the feature flag was modified.
    :ivar str etag:
        The ETag contains a value that you can use to perform operations.
    :ivar dict {string, FeatureFilter[]} conditions:
        Dictionary that contains client_filters List (and server_filters List in future)
    '''

    def __init__(self,
                 key,
                 label=None,
                 state=None,
                 description=None,
                 conditions=None,
                 locked=None,
                 last_modified=None):
        self.key = key
        self.label = label
        self.state = state.name.lower()
        self.description = description
        self.conditions = conditions
        self.last_modified = last_modified
        self.locked = locked

    def __repr__(self):
        featureflag = {
            "Key": self.key,
            "Label": self.label,
            "State": self.state,
            "Locked": self.locked,
            "Description": self.description,
            "Last Modified": self.last_modified,
            "Conditions": custom_serialize_conditions(self.conditions)
        }

        return json.dumps(featureflag, indent=2, ensure_ascii=False)


class FeatureFilter:
    '''
    Feature filters class.

    :ivar str Name:
        Name of the filter
    :ivar dict parameters:
        Name-Value pairs of parameters
    '''

    def __init__(self,
                 name,
                 parameters=None):
        self.name = name
        self.parameters = parameters

    def __repr__(self):
        featurefilter = {
            "name": self.name,
            "parameters": self.parameters
        }
        return json.dumps(featurefilter, indent=2, ensure_ascii=False)

# Feature Flag Helper Functions #


def custom_serialize_conditions(conditions_dict):
    '''
        Helper Function to serialize Conditions

        Args:
            conditions_dict - Dictionary of {str, List[FeatureFilter]}

        Return:
            JSON serializable Dictionary
    '''
    featurefilterdict = {}

    for key, value in conditions_dict.items():
        featurefilters = []
        for featurefilter in value:
            featurefilters.append(str(featurefilter))
        featurefilterdict[key] = featurefilters
    return featurefilterdict


def map_featureflag_to_keyvalue(featureflag):
    '''
        Helper Function to convert FeatureFlag object to KeyValue object

        Args:
            featureflag - FeatureFlag object to be converted

        Return:
            KeyValue object
    '''
    try:
        enabled = False
        if featureflag.state in ("on", "conditional"):
            enabled = True

        feature_flag_value = FeatureFlagValue(id_=featureflag.key,
                                              description=featureflag.description,
                                              enabled=enabled,
                                              conditions=featureflag.conditions)

        set_kv = KeyValue(key=FeatureFlagConstants.FEATURE_FLAG_PREFIX + featureflag.key,
                          label=featureflag.label,
                          value=json.dumps(feature_flag_value,
                                           default=lambda o: o.__dict__,
                                           ensure_ascii=False),
                          content_type=FeatureFlagConstants.FEATURE_FLAG_CONTENT_TYPE,
                          tags={})

        set_kv.locked = featureflag.locked
        set_kv.last_modified = featureflag.last_modified

    except ValueError as exception:
        error_msg = "Exception while converting feature flag to key value: {0}\n{1}".format(featureflag.key, exception)
        raise ValueError(error_msg)

    except Exception as exception:
        error_msg = "Exception while converting feature flag to key value: {0}\n{1}".format(featureflag.key, exception)
        raise Exception(error_msg)

    return set_kv


def map_keyvalue_to_featureflag(keyvalue, show_conditions=True):
    '''
        Helper Function to convert KeyValue object to FeatureFlag object for display

        Args:
            keyvalue - KeyValue object to be converted
            show_conditions - Boolean for controlling whether we want to display "Conditions" or not

        Return:
            FeatureFlag object
    '''
    feature_name = keyvalue.key[len(FeatureFlagConstants.FEATURE_FLAG_PREFIX):]
    feature_flag_value = map_keyvalue_to_featureflagvalue(keyvalue)
    state = FeatureState.OFF
    if feature_flag_value.enabled:
        state = FeatureState.ON

    conditions = feature_flag_value.conditions

    # if conditions["client_filters"] list is not empty, make state conditional
    filters = conditions["client_filters"]

    if filters and state == FeatureState.ON:
        state = FeatureState.CONDITIONAL

    feature_flag = FeatureFlag(feature_name,
                               keyvalue.label,
                               state,
                               feature_flag_value.description,
                               conditions,
                               keyvalue.locked,
                               keyvalue.last_modified)

    # By Default, we will try to show conditions unless the user has
    # specifically filtered them using --fields arg.
    # But in some operations like 'Delete feature', we don't want
    # to display all the conditions as a result of delete operation
    if not show_conditions:
        del feature_flag.conditions
    return feature_flag


def map_keyvalue_to_featureflagvalue(keyvalue):
    '''
        Helper Function to convert value string to a valid FeatureFlagValue.
        Throws Exception if value is an invalid JSON.

        Args:
            keyvalue - KeyValue object

        Return:
            Valid FeatureFlagValue object
    '''

    try:
        # Make sure value string is a valid json
        feature_flag_dict = shell_safe_json_parse(keyvalue.value)
        feature_name = keyvalue.key[len(FeatureFlagConstants.FEATURE_FLAG_PREFIX):]

        # Make sure value json has all the fields we support in the backend
        valid_fields = {
            'id',
            'description',
            'enabled',
            'conditions'}
        if valid_fields != feature_flag_dict.keys():
            logger.debug("'%s' feature flag is missing required values or it contains ", feature_name +
                         "unsupported values. Setting missing value to defaults and ignoring unsupported values\n")

        conditions = feature_flag_dict.get('conditions', None)
        if conditions:
            client_filters = conditions.get('client_filters', [])

            # Convert all filters to FeatureFilter objects
            client_filters_list = []
            for client_filter in client_filters:
                # If there is a filter, it should always have a name
                # In case it doesn't, ignore this filter
                name = client_filter.get('name')
                if name:
                    params = client_filter.get('parameters', {})
                    client_filters_list.append(FeatureFilter(name, params))
                else:
                    logger.warning("Ignoring this filter without the 'name' attribute:\n%s",
                                   json.dumps(client_filter, indent=2, ensure_ascii=False))
            conditions['client_filters'] = client_filters_list

        feature_flag_value = FeatureFlagValue(id_=feature_name,
                                              description=feature_flag_dict.get(
                                                  'description', ''),
                                              enabled=feature_flag_dict.get(
                                                  'enabled', False),
                                              conditions=conditions)

    except ValueError as exception:
        error_msg = "Invalid value. Unable to decode the following JSON value: \n" +\
                    "key:{0} value:{1}\nFull exception: \n{2}".format(keyvalue.key, keyvalue.value, str(exception))
        raise ValueError(error_msg)

    except:
        logger.error("Exception while parsing feature flag. key:%s value:%s.", keyvalue.key, keyvalue.value)
        raise

    return feature_flag_value
